import requests

from urlparse import parse_qsl, urlparse

from django.db.models import Max
from django.utils.text import slugify

from bs4 import BeautifulSoup
from dateutil.parser import parse

from tournamentcontrol.competition.models import Club, Match


def _next(manager, attr='order'):
    return (manager.aggregate(next=Max(attr))['next'] or 0) + 1


def sync(division):
    assert division.sportingpulse_url

    # Try to fetch the data from the specified SportingPulse URL
    source = requests.get(division.sportingpulse_url)
    assert source.status_code == 200

    ground_index = {}
    for venue in division.season.venues.select_related('grounds'):
        for ground in venue.grounds.all():
            ground_index[ground.title] = ground.pk

    stage_pool_index = {}
    for stage in division.stages.select_related('pools'):
        if stage.pools.count():
            for pool in stage.pools.all():
                stage_pool_index[pool.title] = (stage.pk, pool.pk)
        else:
            stage_pool_index[stage.title] = (stage.pk, None)

    # Use BeautifulSoup to construct a parse tree for data scraping
    soup = BeautifulSoup(source.content)

    # Initial values for content markers
    group = None
    round = None

    for row in soup.findAll('tr'):
        label = None

        # Fetch all the table data cells
        cols = row.findAll('td')

        # There are 2 linked fields, Venue and Match detail (in that order)
        links = row.findAll('a')

        # Marks the start of a Stage or StageGroup
        if len(cols) == 1:
            g = cols[0].text.strip()
            stage, group = stage_pool_index[g]

        elif cols:
            # Last link on the row holds the metadata for the match in the
            # external system.
            meta = dict(parse_qsl(urlparse(links[-1].attrs['href']).query))
            external_identifier = meta['fixture']

            # Round number & label
            r = cols[0].text.strip()
            l = cols[1].findAll('i')
            if r:
                round = r
            if l:
                label = l[0].text.strip()[1:-1]

            # Date, time, and place
            dt = parse("%s %s" % (cols[1].text.strip()[:8],
                                  cols[2].text.strip()), dayfirst=True)
            play_at = ground_index.get(cols[3].text.strip())

            # Teams & scores
            home_team_score = cols[4].text.strip()
            home_team = cols[5].text.strip()
            away_team = cols[7].text.strip()
            away_team_score = cols[8].text.strip()

            try:
                home_club = division.season.competition.clubs.get(
                    title=home_team)
            except Club.DoesNotExist:
                home_club = None

            try:
                away_club = division.season.competition.clubs.get(
                    title=away_team)
            except Club.DoesNotExist:
                away_club = None

            home_team, __ = division.teams.get_or_create(
                title=home_team, defaults={'slug': slugify(home_team),
                                           'club': home_club,
                                           'order': _next(division.teams)})

            away_team, __ = division.teams.get_or_create(
                title=away_team, defaults={'slug': slugify(away_team),
                                           'club': away_club,
                                           'order': _next(division.teams)})

            defaults = {
                'stage_id': stage,
                'stage_group_id': group,
                'round': round,
                'label': label,
                'date': dt.date(),
                'time': dt.time(),
                'play_at_id': play_at,
                'home_team': home_team,
                'home_team_score': home_team_score or None,
                'away_team': away_team,
                'away_team_score': away_team_score or None,
            }

            # Create or Update the match record based on scraped data.
            match, created = Match.objects.get_or_create(
                external_identifier=external_identifier, defaults=defaults)

            if not created:
                for k, v in defaults.items():
                    setattr(match, k, v)

            match.clean()
            match.save()



from django.core.urlresolvers import reverse_lazy
from django.utils.functional import cached_property


class AdminUrlMixin(object):
    """
    Mixin which can be used to add API to a model for use by standard admin
    component templates, tags, and filters.
    """

    def _get_admin_namespace(self):
        raise NotImplementedError

    def _get_url_args(self):
        raise NotImplementedError

    @cached_property
    def urls(self):
        namespace = self._get_admin_namespace()
        args = self._get_url_args()
        crud = {
            'edit': reverse_lazy(
                '%s:edit' % namespace, args=args),
            'delete': reverse_lazy(
                '%s:delete' % namespace, args=args),
            'perms': reverse_lazy(
                '%s:perms' % namespace, args=args),
        }
        return crud



from django.conf.urls import include, url

from touchtechnology.common.sites import AccountsSite
from tournamentcontrol.competition.sites import competition

accounts = AccountsSite()

urlpatterns = [
    url(r'^accounts/', include(accounts.urls)),
    url(r'^', include(competition.urls)),
]



from django.conf.urls import include, url
from example_app import sites, views

from touchtechnology.common.sites import AccountsSite
from touchtechnology.news.sites import NewsSite

test_context_processors = sites.TestContextProcessorsSite()
test_date_time_field = sites.TestDateTimeFieldSite()
test_generic_views = sites.TestGenericViewsSite()
test_query_string = sites.TestQueryStringSite()
test_pagination = sites.TestPaginationSite()

accounts = AccountsSite()
news = NewsSite()

urlpatterns = [
    url(r'^context-processors/', include(test_context_processors.urls)),
    url(r'^date-time-field/', include(test_date_time_field.urls)),
    url(r'^generic/', include(test_generic_views.urls)),
    url(r'^query-string/', include(test_query_string.urls)),
    url(r'^pagination/', include(test_pagination.urls)),
    url(r'^accounts/', include(accounts.urls)),  # can't login without it
    url(r'^news/', include(news.urls)),
    url(r'^', include('touchtechnology.common.urls')),
    url(r'^$', views.index, name='index'),
]



from decimal import Decimal
from unittest import skipIf

import six
from test_plus import TestCase

from touchtechnology.common.forms import boolean_choice_field_coerce
from touchtechnology.common.utils import average, get_mod_func, timezone_choice


class Average(TestCase):

    def test_integers(self):
        series = (1, 2, 3, 4)
        res = average(series)
        self.assertEqual(res, 2.5)

    def test_floats(self):
        series = (1.25, 1.5, 1.75)
        res = average(series)
        self.assertEqual(res, 1.5)

    @skipIf(six.PY3, "Floating point's seem to have changed?")
    def test_complex_floats(self):
        series = (5.0 / 3, 35.0 / 6, 22.0 / 7)
        res = average(series)
        self.assertEqual(res, Decimal('3.54761904762'))

    def test_decimals(self):
        series = (Decimal(5) / 3, Decimal(35) / 6, Decimal(22) / 7)
        res = average(series)
        self.assertEqual(res, Decimal('3.547619047619047619047619047'))


class BooleanChoiceFieldCoerce(TestCase):

    def test_true_boolean(self):
        result = boolean_choice_field_coerce(True)
        self.assertTrue(result)

    def test_false_boolean(self):
        result = boolean_choice_field_coerce(False)
        self.assertFalse(result)

    def test_true_integer(self):
        result = boolean_choice_field_coerce('1')
        self.assertTrue(result)

    def test_false_integer(self):
        result = boolean_choice_field_coerce('0')
        self.assertFalse(result)

    def test_true_string(self):
        result = boolean_choice_field_coerce(1)
        self.assertTrue(result)

    def test_false_string(self):
        result = boolean_choice_field_coerce(0)
        self.assertFalse(result)


class GetModFunc(TestCase):

    def test_get_mod_func(self):
        mod_name, func_name = get_mod_func(
            'django.contrib.auth.tokens.default_token_generator')
        self.assertEqual(mod_name, 'django.contrib.auth.tokens')
        self.assertEqual(func_name, 'default_token_generator')

    def test_no_dot(self):
        mod_name, func_name = get_mod_func('django')
        self.assertEqual(mod_name, 'django')
        self.assertEqual(func_name, '')


class TimeZoneChoice(TestCase):

    def test_one_part(self):
        self.assertEqual(
            timezone_choice(u'UTC'),
            (u'UTC', u'UTC'))

    def test_two_parts(self):
        self.assertEqual(
            timezone_choice(u'Australia/Sydney'),
            (u'Australia/Sydney', u'Sydney'))

    def test_three_parts(self):
        self.assertEqual(
            timezone_choice(u'America/Indiana/Indianapolis'),
            (u'America/Indiana/Indianapolis', u'Indianapolis (Indiana)'))



from django.contrib.sitemaps import GenericSitemap


class CustomGenericSitemap(GenericSitemap):
    """
    Extend the provided GenericSitemap so that we can also pass in the
    Application instance. This then allows us to use the .reverse method to
    determine the right URL for the object with regard to the object.
    """
    def __init__(self, info_dict, *args, **kwargs):
        super(CustomGenericSitemap, self).__init__(info_dict, *args, **kwargs)
        self.app = info_dict['app']


class SeasonSitemap(CustomGenericSitemap):
    def location(self, obj):
        return self.app.reverse('season', kwargs={
            'competition': obj.competition.slug,
            'season': obj.slug})


class DivisionSitemap(CustomGenericSitemap):
    def location(self, obj):
        return self.app.reverse('division', kwargs={
            'competition': obj.season.competition.slug,
            'season': obj.season.slug,
            'division': obj.slug})


class MatchSitemap(CustomGenericSitemap):
    def items(self):
        "Only return matches that have known opponents, will exclude Byes"
        return self.queryset.filter(home_team__isnull=False,
                                    away_team__isnull=False)

    def location(self, obj):
        return self.app.reverse('match', kwargs={
            'competition': obj.stage.division.season.competition.slug,
            'season': obj.stage.division.season.slug,
            'division': obj.stage.division.slug,
            'match': obj.pk})



from celery import shared_task

from tournamentcontrol.competition import sportingpulse
from tournamentcontrol.competition.decorators import competition_by_slug
from tournamentcontrol.competition.utils import generate_scorecards


@shared_task
def generate_pdf_scorecards(matches, templates, extra_context, stage=None,
                            **kwargs):
    return generate_scorecards(
        matches, templates, 'pdf', extra_context, stage, **kwargs)


@shared_task
@competition_by_slug
def synchronise_with_sportingpulse(season=None, **kwargs):
    for division in season.divisions.filter(sportingpulse_url__isnull=False):
        sportingpulse.sync(division)
        # Issue with ladders, so force all matches to be re-saved and force a
        # ladder recalculation.
        for match in division.matches.all():
            match.save()



import os.path
import re

from dateutil.parser import parse as parse_datetime
from django.test.utils import modify_settings, override_settings
from django.utils import timezone
from test_plus import TestCase

from touchtechnology.common.tests.factories import UserFactory
from touchtechnology.news.tests.factories import ArticleFactory

TEST_DIR = os.path.join(os.path.dirname(__file__), 'html')


def get_html_fragment(filename):
    with open(os.path.join(TEST_DIR, filename)) as fp:
        html = fp.read()
    return html


class PublishedTimeZoneTestCase(TestCase):

    def setUp(self):
        self.superuser = UserFactory.create(is_staff=True, is_superuser=True)
        self.article = ArticleFactory.create(
            published=parse_datetime('2013-1-1 12:00+00:00'))

    def test_published_datetime_html(self):
        self.assertLoginRequired('admin:news:article:edit', self.article.pk)
        with self.login(self.superuser):
            self.get('admin:news:article:edit', self.article.pk)
            for i in range(5):
                html = get_html_fragment(
                    'select_date_time_published_%d.html' % i)
                self.assertInHTML(html, self.last_response.content, count=1)

    def test_published_datetime_html_tz(self):
        with timezone.override('Australia/Sydney'):
            with self.login(self.superuser):
                self.get('admin:news:article:edit', self.article.pk)
                html = get_html_fragment(
                    'select_date_time_published_3-Sydney.html')
                self.assertInHTML(html, self.last_response.content, count=1)



from django.conf import settings
from django.utils.module_loading import import_string

__all__ = (
    'APP_ROUTING',
    'AUTH_FORM_CLASS',
    'CURRENCY_ABBREVIATION',
    'CURRENCY_SYMBOL',
    'PAGINATE_BY',
    'SITEMAP_CACHE_DURATION',
    'SITEMAP_EDIT_PARENT',
    'SITEMAP_HTTPS_OPTION',
    'SITEMAP_ROOT',
    'STORAGE_FOLDER',
    'STORAGE_URL',
)

A = lambda n, d: getattr(settings, 'AUTHENTICATION_' + n, d)
C = lambda n, d: import_string(A(n, d))
S = lambda n, d=None: getattr(settings, 'TOUCHTECHNOLOGY_' + n, d)

APP_ROUTING = S('APP_ROUTING', ())
AUTH_FORM_CLASS = C('FORM_CLASS',
                    'django.contrib.auth.forms.AuthenticationForm')
CURRENCY_ABBREVIATION = S('CURRENCY_ABBREVIATION', 'AUD')
CURRENCY_SYMBOL = S('CURRENCY_SYMBOL', '$')
PAGINATE_BY = S('PAGINATE_BY', 5)
SITEMAP_CACHE_DURATION = S('SITEMAP_DURATION')
SITEMAP_EDIT_PARENT = S('SITEMAP_EDIT_PARENT', False)
SITEMAP_HTTPS_OPTION = S('HTTPS_OPTION', False)
SITEMAP_ROOT = S('SITEMAP_ROOT')
STORAGE_FOLDER = S('STORAGE_FOLDER')
STORAGE_URL = S('STORAGE_URL')



import json
import logging
import os.path
import re
from datetime import date, datetime, time
from operator import itemgetter

import pytz
import six
from django import forms
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.forms import AuthenticationForm
from django.contrib.auth.models import Group
from django.contrib.contenttypes.models import ContentType
from django.forms.formsets import formset_factory
from django.forms.utils import flatatt
from django.forms.widgets import (
    CheckboxSelectMultiple,
    CheckboxFieldRenderer,
    ChoiceFieldRenderer,
    SelectMultiple,
    MultiWidget as MWB,
)
from django.utils import timezone
from django.utils.encoding import smart_str
from django.utils.functional import cached_property
from django.utils.html import format_html
from django.utils.http import urlencode
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _
from froala_editor.widgets import FroalaEditor
from guardian.models import GroupObjectPermission, UserObjectPermission
from guardian.shortcuts import (
    assign_perm,
    remove_perm,
)

from touchtechnology.common.mixins import BootstrapFormControlMixin
from touchtechnology.common.utils import timezone_choices

logger = logging.getLogger(__name__)


def groups_widget(f):
    return forms.CheckboxSelectMultiple(choices=f.choices)


class MultiWidget(MWB):

    def build_attrs(self, *args, **kwargs):
        """
        In the underlying render process for the widget, build_attrs will be
        used to determine the ``final_attrs`` to be associated to each widget.

        Unfortunately in Django 1.10 it seems that "required" is being added
        and this is producing different rendered output than that of earlier
        releases.

        This method is used to restore the pre-1.10 behaviour so our tests can
        continue passing across versions 1.8-1.10.
        """
        import django
        attrs = super(MultiWidget, self).build_attrs(*args, **kwargs)
        if django.VERSION >= (1, 10):
            attrs.pop('required', None)
        return attrs


class RadioChoiceInput(forms.widgets.RadioChoiceInput):
    def render(self, name=None, value=None, attrs=None, choices=()):
        if self.id_for_label:
            label_for = format_html(' for="{0}"', self.id_for_label)
        else:
            label_for = ''
        # The special sauce is to use smart_str, on tracing it was found that
        # format_html was returning SafeBytes whereas the implementation on the
        # super class was returning SafeText.
        return smart_str(
            format_html('<label{0} class="radio-inline">{1} {2}</label>',
                        label_for, self.tag(), self.choice_label))

    def tag(self):
        final_attrs = dict(self.attrs, type=self.input_type,
                           name=self.name, value=self.choice_value)
        if self.is_checked():
            final_attrs['checked'] = 'checked'
        return format_html('<input{0} />', flatatt(final_attrs))


class RadioFieldRenderer(forms.widgets.RadioFieldRenderer):
    choice_input_class = RadioChoiceInput

    def render(self):
        self.attrs.setdefault('class', 'ui-icheck')
        start_tag = '<div class="icheck">'
        output = [start_tag]
        for i, choice in enumerate(self.choices):
            w = self.choice_input_class(self.name, self.value,
                                        self.attrs.copy(), choice, i)
            output.append(smart_str(w))
        output.append('</div>')
        return mark_safe('\n'.join(output))


class LabelFromInstanceMixin(object):
    def __init__(self, label_from_instance='name', *args, **kwargs):
        super(LabelFromInstanceMixin, self).__init__(*args, **kwargs)
        self._label_from_instance = label_from_instance

    def label_from_instance(self, obj):
        if isinstance(self._label_from_instance, six.string_types):
            value = getattr(obj, self._label_from_instance)
        elif callable(self._label_from_instance):
            value = self._label_from_instance(obj)
        else:
            value = obj
        if callable(value):
            try:
                value = value(obj)
            except TypeError:
                value = value()
        return smart_str(value)


def boolean_choice_field_coerce(value):
    return bool(int(value))


class BooleanSelect(forms.RadioSelect):
    renderer = RadioFieldRenderer

    def render(self, name, value, *args, **kwargs):
        value = int(value or '0')
        return super(BooleanSelect, self).render(name, value, *args, **kwargs)


class BooleanChoiceField(forms.TypedChoiceField):
    widget = BooleanSelect

    def __init__(self, positive_label=_('Yes'), negative_label=_('No'),
                 *args, **kwargs):
        super(BooleanChoiceField, self).__init__(
            choices=((1, positive_label), (0, negative_label)),
            coerce=boolean_choice_field_coerce,
            *args, **kwargs)


class EmailField(forms.EmailField):
    """
    Custom EmailField which will transform email address to lowercase unless
    advised otherwise as this SHOULD NOT be significant.
    """
    def __init__(self, lowercase=True, *args, **kwargs):
        super(EmailField, self).__init__(*args, **kwargs)
        self.lowercase = lowercase

    def clean(self, data):
        data = super(EmailField, self).clean(data)
        if self.lowercase:
            data = data.lower()
        return data


class ModelMultipleChoiceField(LabelFromInstanceMixin,
                               forms.ModelMultipleChoiceField):
    widget = forms.CheckboxSelectMultiple

    def __init__(self, select_related=True, *args, **kwargs):
        super(ModelMultipleChoiceField, self).__init__(*args, **kwargs)
        if select_related and self.queryset is not None:
            self.queryset = self.queryset.select_related()


class EmailAuthenticationForm(BootstrapFormControlMixin, AuthenticationForm):
    username = EmailField(label=_('Email Address'))

    def __init__(self, *args, **kwargs):
        super(EmailAuthenticationForm, self).__init__(*args, **kwargs)
        for field_name in self.fields:
            self.fields[field_name].widget.attrs['class'] += ' text-center'


class HTMLWidget(FroalaEditor):

    def trigger_froala(self, el_id, options):
        """
        We don't want to do inline scripts, we'll handle the loading once from
        froala.js, so return an empty string to overload the base FroalaEditor
        implementation.
        """
        return u''

    class Media:
        css = {
            'all': (
                'https://cdnjs.cloudflare.com/ajax/libs/codemirror/5.3.0/'
                'codemirror.min.css',
                'touchtechnology/common/css/froala_themes/dark.min.css',
                'touchtechnology/common/css/froala_themes/gray.min.css',
                'touchtechnology/common/css/froala_themes/red.min.css',
                'touchtechnology/common/css/froala_themes/royal.min.css',
            ),
        }
        js = (
            'https://cdnjs.cloudflare.com/ajax/libs/codemirror/5.3.0/'
            'codemirror.min.js',
            'touchtechnology/common/js/froala.js',
        )


class HTMLField(forms.CharField):
    widget = HTMLWidget

    def widget_attrs(self, widget):
        options = getattr(settings, 'FROALA_EDITOR_OPTIONS', {})
        attrs = {
            'data-options': json.dumps(options),
        }
        return attrs

    def clean(self, value):
        if isinstance(value, six.string_types):
            value = value.encode('ascii', 'xmlcharrefreplace')
        return super(HTMLField, self).clean(value)


class GoogleMapsWidget(MultiWidget):
    def __init__(self, height, width, zoom, attrs=None):
        self.attrs = attrs or {}
        self.height = height
        self.width = width
        self.zoom = zoom
        widgets = (
            forms.HiddenInput(attrs=attrs),
            forms.HiddenInput(attrs=attrs),
            forms.HiddenInput(attrs=attrs),
        )
        super(GoogleMapsWidget, self).__init__(widgets, attrs)

    def decompress(self, value):
        if value:
            return value.split(',')
        return ('', '', '')

    def render(self, *args, **kwargs):
        res = super(GoogleMapsWidget, self).render(*args, **kwargs)
        return u'<div class="mapwidget">' + res + u'</div>'

    class Media:
        css = {
            'all': ('touchtechnology/common/css/location_widget.css',)
        }
        js = (
            '//maps.googleapis.com/maps/api/js?v=3',
            'touchtechnology/common/js/location_widget.js',
        )


class BootstrapGoogleMapsWidget(GoogleMapsWidget):
    def render(self, *args, **kwargs):
        base = super(BootstrapGoogleMapsWidget, self).render(*args, **kwargs)
        return u'''
            <div class="form-group">
                <label class="control-label col-sm-3">%s</label>
                <div class="col-sm-7">%s</div>
            </div>
        ''' % (_("Location"), base)


class GoogleMapsField(forms.MultiValueField):
    def __init__(self, max_length, width=300, height=200, zoom=8,
                 *args, **kwargs):
        fields = (
            forms.CharField(max_length=25, required=False),
            forms.CharField(max_length=25, required=False),
            forms.IntegerField(required=False),
        )
        kwargs['widget'] = BootstrapGoogleMapsWidget(height, width, zoom)
        super(GoogleMapsField, self).__init__(fields, *args, **kwargs)

    def compress(self, data_list):
        return ','.join([smart_str(d) for d in data_list])

    def clean(self, data):
        if [d for d in data if d]:
            return ','.join(data)
        return super(GoogleMapsField, self).clean(data)


DAY_CHOICES = [('', '')] + list(zip(*[range(1, 32)] * 2))

MONTH_CHOICES = (
    ('', ''),
    (1, 'January'), (2, 'February'), (3, 'March'), (4, 'April'), (5, 'May'),
    (6, 'June'), (7, 'July'), (8, 'August'), (9, 'September'), (10, 'October'),
    (11, 'November'), (12, 'December'),
)

HOUR_CHOICES = [('', '')] + list(zip(*[range(0, 24, 1)] * 2))

MINUTES = range(0, 60)
MINUTE_CHOICES = [('', '')] + list(zip(MINUTES, ['%02d' % m for m in MINUTES]))

TIMEZONE_CHOICES = list(timezone_choices(
    pytz.country_names, pytz.country_timezones, itemgetter(1)))

# Add UTC and MGT
TIMEZONE_CHOICES.insert(1, (_('Universal'), [
    ('UTC', _("Coordinated Universal Time")),
    ('GMT', _("Greenwich Mean Time")),
]))


class SelectDateWidget(MultiWidget):
    def __init__(self, attrs=None):
        widgets = (
            forms.Select(choices=DAY_CHOICES, attrs=attrs),
            forms.Select(choices=MONTH_CHOICES, attrs=attrs),
            forms.TextInput(attrs=attrs),
        )
        super(SelectDateWidget, self).__init__(widgets, attrs)

    def decompress(self, value):
        if value:
            return (value.day, value.month, value.year)
        return (None, None, None)

    def format_output(self, rendered_widgets):
        return " ".join(rendered_widgets)


class SelectDateHiddenWidget(SelectDateWidget):
    def __init__(self, attrs=None):
        widgets = (
            forms.HiddenInput(attrs=attrs),
            forms.HiddenInput(attrs=attrs),
            forms.HiddenInput(attrs=attrs),
        )
        super(SelectDateWidget, self).__init__(widgets, attrs)


class SelectDateField(forms.MultiValueField):
    widget = SelectDateWidget
    hidden_widget = SelectDateHiddenWidget

    def __init__(self, *args, **kwargs):
        fields = (
            forms.IntegerField(required=False),
            forms.IntegerField(required=False),
            forms.IntegerField(required=False),
        )
        super(SelectDateField, self).__init__(fields, *args, **kwargs)

    def compress(self, data_list):
        return date(data_list[2], data_list[1], data_list[0])

    def clean(self, data):
        if data is None:
            data = ['', '', '']

        if not [d for d in data if d]:
            if self.required:
                raise forms.ValidationError('This field is required.')
            return None

        try:
            day, month, year = map(int, data)
            d = date(year, month, day)
        except ValueError:
            raise forms.ValidationError('Please enter a valid date.')
        return d


class SelectTimeWidget(MultiWidget):
    def __init__(self, attrs=None):
        widgets = (
            forms.Select(choices=HOUR_CHOICES, attrs=attrs),
            forms.Select(choices=MINUTE_CHOICES, attrs=attrs),
        )
        super(SelectTimeWidget, self).__init__(widgets, attrs)

    def decompress(self, value):
        if value is not None:
            return (value.hour, value.minute)
        return (None, None)

    def format_output(self, rendered_widgets):
        return " ".join(rendered_widgets)


class SelectTimeHiddenWidget(SelectTimeWidget):
    def __init__(self, attrs=None):
        widgets = (
            forms.HiddenInput(attrs=attrs),
            forms.HiddenInput(attrs=attrs),
        )
        super(SelectTimeWidget, self).__init__(widgets, attrs)


class SelectTimeField(forms.MultiValueField):
    widget = SelectTimeWidget
    hidden_widget = SelectTimeHiddenWidget

    def __init__(self, *args, **kwargs):
        fields = (
            forms.IntegerField(required=False),
            forms.IntegerField(required=False),
        )
        super(SelectTimeField, self).__init__(fields, *args, **kwargs)

    def compress(self, data_list):
        return time(data_list[0], data_list[1])

    def clean(self, data):
        if not data:
            if self.required:
                raise forms.ValidationError('This field is required.')
            data = ('', '')

        t = None

        try:
            try:
                try:
                    hour, minute = data
                except ValueError:
                    raise forms.ValidationError(repr(data))

                try:
                    hour = int(hour)
                    assert 0 <= hour <= 23
                except (AssertionError, ValueError):
                    raise forms.ValidationError('Hour must be in 0..23')

                try:
                    minute = int(minute)
                    assert 0 <= minute <= 59
                except (AssertionError, ValueError):
                    raise forms.ValidationError('Minute must be in 0..59')

                try:
                    t = time(hour, minute)
                except ValueError as e:
                    raise forms.ValidationError(smart_str(e).capitalize())

            except TypeError as e:
                raise forms.ValidationError('Please enter a valid time.')

        except forms.ValidationError:
            if self.required or any(data):
                raise

        return t


class SelectDateTimeWidget(MultiWidget):
    def __init__(self, attrs=None):
        widgets = (
            forms.Select(choices=DAY_CHOICES, attrs=attrs),
            forms.Select(choices=MONTH_CHOICES, attrs=attrs),
            forms.TextInput(attrs=attrs),
            forms.Select(choices=HOUR_CHOICES, attrs=attrs),
            forms.Select(choices=MINUTE_CHOICES, attrs=attrs),
        )
        if settings.USE_TZ:
            widgets += (
                forms.Select(choices=TIMEZONE_CHOICES, attrs=attrs),
            )
        super(SelectDateTimeWidget, self).__init__(widgets, attrs)

    def decompress(self, value):
        if settings.USE_TZ:
            tzinfo = timezone.get_current_timezone()
            tzname = timezone.get_current_timezone_name()
            logger.debug('USE_TZ is True, timezone is %s (%s)',
                         tzinfo, tzname)

            if value:
                logger.debug('Value is set (%s)', value)
                if timezone.is_naive(value):
                    value = timezone.make_aware(
                        value, timezone.get_default_timezone())
                    logger.debug('Naive value, set default timezone (%s)',
                                 value)
                value = value.astimezone(tzinfo)
                logger.debug('Ensure timezone is correctly set to %s (%s)',
                             tzname, value)
                values = (value.day, value.month, value.year,
                          value.hour, value.minute, tzname)
            else:
                values = (None, None, None, None, None, None)

        elif value:
            values = (value.day, value.month, value.year,
                      value.hour, value.minute)

        else:
            values = (None, None, None, None, None)

        return values

    def format_output(self, rendered_widgets):
        date_part = " ".join(rendered_widgets[:3])
        time_part = " ".join(rendered_widgets[3:])
        output = """<span class="date_part">%(date_widget)s</span>
        <span class="time_part%(timezone)s">%(time_widget)s</span>""" % {
            'date_widget': date_part,
            'time_widget': time_part,
            'timezone': ' tz' if settings.USE_TZ else '',
        }
        return output


class SelectDateTimeHiddenWidget(SelectDateTimeWidget):
    def __init__(self, attrs=None):
        widgets = (
            forms.HiddenInput(),
            forms.HiddenInput(),
            forms.HiddenInput(),
            forms.HiddenInput(),
            forms.HiddenInput(),
        )
        if settings.USE_TZ:
            widgets += (
                forms.HiddenInput(),
            )
        super(SelectDateTimeWidget, self).__init__(widgets, attrs)


class SelectDateTimeField(forms.MultiValueField):
    widget = SelectDateTimeWidget
    hidden_widget = SelectDateTimeHiddenWidget

    def __init__(self, *args, **kwargs):
        fields = (
            forms.IntegerField(required=False),
            forms.IntegerField(required=False),
            forms.IntegerField(required=False),
            forms.IntegerField(required=False),
            forms.IntegerField(required=False),
        )
        if settings.USE_TZ:
            fields += (
                forms.CharField(required=True),
            )
        super(SelectDateTimeField, self).__init__(fields, *args, **kwargs)

    def compress(self, data_list):
        return date(data_list[2], data_list[1], data_list[0],
                    data_list[3], data_list[4])

    def clean(self, data):
        if data is None:
            data = ['', '', '', '', '']

        if not [d for d in data if d]:
            if self.required:
                raise forms.ValidationError('This field is required.')
            return None

        try:
            day, month, year, hour, minute = map(int, data[:5])
            value = datetime(year, month, day, hour, minute)
            if settings.USE_TZ:
                try:
                    tzinfo = pytz.timezone(data[5])
                except IndexError:
                    raise forms.ValidationError('Please select a time zone.')
                except pytz.exceptions.UnknownTimeZoneError:
                    raise forms.ValidationError('Please select a valid time '
                                                'zone.')
                value = timezone.make_aware(value, tzinfo)
        except ValueError:
            raise forms.ValidationError('Please enter a valid date and time.')

        return value


class TemplateChoiceIterator(object):
    def __init__(self, field):
        self.field = field

    def __iter__(self):
        folder = os.path.join(self.field.template_base,
                              self.field.template_folder)
        choices = []
        if self.field.recursive:
            for root, dirs, files in os.walk(folder):
                for f in files:
                    if self.field.match is None or \
                       self.field.match_re.search(f):
                        f = os.path.join(root, f)
                        base = f.replace(self.field.template_base, '')
                        choices.append((base, f.replace(folder, '', 1)))
        else:
            try:
                for f in os.listdir(folder):
                    full_file = os.path.join(folder, f)
                    if os.path.isfile(full_file) and \
                            (self.field.match is None or
                             self.field.match_re.search(f)):
                        choices.append((full_file, f))
            except OSError:
                pass

        if not self.field.required:
            yield (u"", self.field.empty_label)

        if choices:
            yield ('Static template', choices)

    def choice(self, obj):
        path = obj.path.replace(self.field.template_folder, '', 1)
        return (obj.path, '%s (%s)' % (path, obj.name))


class TemplatePathFormField(forms.ChoiceField):
    def __init__(self, template_base, template_folder, match=None,
                 recursive=False, required=True, widget=None, label=None,
                 initial=None, help_text=None, cache_choices=False,
                 empty_label=_(u"Default"),
                 # This will never be used, but it's now a CharField in the
                 # database so we need to pull it out of the keyword
                 # arguments.
                 max_length=None,
                 *args, **kwargs):

        super(TemplatePathFormField, self).__init__(required=required,
                                                    widget=widget, label=label,
                                                    initial=initial,
                                                    help_text=help_text,
                                                    *args, **kwargs)

        self.template_base = template_base
        self.template_folder = template_folder
        self.match = match
        self.recursive = recursive
        self.empty_label = empty_label

        if self.match is not None:
            self.match_re = re.compile(self.match)

        if not self.template_base.endswith('/'):
            self.template_base += '/'

        self.widget.choices = self.choices

    def _get_choices(self):
        return TemplateChoiceIterator(self)

    choices = property(_get_choices, forms.ChoiceField._set_choices)


class SelectTimezoneForm(forms.Form):
    """
    Used by touchtechnology.common.context_processors.tz to generate a
    dropdown list of timezones to be used with the TimeZone middleware and
    set-timezone view.
    """
    timezone = forms.ChoiceField(choices=TIMEZONE_CHOICES,
                                 initial=timezone.get_current_timezone)


class UserMixin(BootstrapFormControlMixin):
    """
    Mixin class to be used in forms that need a `User` object passed in.
    """
    def __init__(self, user=None, *args, **kwargs):
        self.user = user
        super(UserMixin, self).__init__(*args, **kwargs)


class SuperUserSlugMixin(UserMixin):
    """
    Mixin class that will attempt to remove a field named `slug` from the
    form if the user cannot be identified, or if they are not a superuser.

    Handy for allowing advanced users to have finer control over the slug
    of a page, while providing default behaviour for regular users.
    """
    def __init__(self, *args, **kwargs):
        super(SuperUserSlugMixin, self).__init__(*args, **kwargs)
        if self.user is None or not self.user.is_superuser:
            self.fields.pop('slug', None)
            self.fields.pop('slug_locked', None)
        else:
            self.fields['slug'].required = False
            self.fields['slug'].help_text = _("If left blank, this will "
                                              "be automatically set based "
                                              "on the title.")


class RedefineModelForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        super(RedefineModelForm, self).__init__(*args, **kwargs)
        for field, _kw in self.Meta.redefine:
            kw = {}
            for key, val in _kw.items():
                if callable(val) and not type(val) == type:
                    kw[key] = val(self.fields[field])
                else:
                    kw[key] = val
            model_field = self.Meta.model._meta.get_field(field)
            self.fields[field] = model_field.formfield(**kw)


class ModelChoiceField(LabelFromInstanceMixin, forms.ModelChoiceField):

    def __init__(self, select_related=True, *args, **kwargs):
        super(ModelChoiceField, self).__init__(*args, **kwargs)
        if select_related and self.queryset is not None:
            self.queryset = self.queryset.select_related()


class SitemapNodeModelChoiceField(ModelChoiceField):
    def __init__(self, label_from_instance='title', *args, **kwargs):
        super(SitemapNodeModelChoiceField, self).__init__(
            label_from_instance=label_from_instance, *args, **kwargs)


class RegistrationForm(forms.Form):
    first_name = forms.CharField(max_length=30)
    last_name = forms.CharField(max_length=30)
    email = EmailField()
    password1 = forms.CharField(widget=forms.PasswordInput,
                                label=_('Password'))
    password2 = forms.CharField(widget=forms.PasswordInput,
                                label=_('Password (again)'))

    def clean_email(self):
        UserModel = get_user_model()
        email = self.cleaned_data.get('email')
        if UserModel.objects.filter(email__iexact=email):
            raise forms.ValidationError(_("This email address is already in "
                                          "use. Please supply a different "
                                          "email address."))
        return email

    def clean_password2(self):
        password1 = self.cleaned_data.get('password1')
        password2 = self.cleaned_data.get('password2')
        if password1 and password2:
            if password1 != password2:
                raise forms.ValidationError(_("The two password fields "
                                              "didn't match."))
        return password2


###################
# django-guardian #
###################

class iCheckFieldRenderer(CheckboxFieldRenderer):
    def render(self):
        """
        Outputs a <div> for this set of choice fields.
        If an id was given to the field, it is applied to the <div> (each item
        in the list will get an id of `$id_$i`).
        """
        id_ = self.attrs.get('id', None)
        if id_:
            start_tag = format_html('<div class="icheck" id="{0}">', id_)
        else:
            start_tag = '<div class="icheck">'
        output = [start_tag]
        for i, choice in enumerate(self.choices):
            choice_value, choice_label = choice
            if isinstance(choice_label, (tuple, list)):
                attrs_plus = self.attrs.copy()
                if id_:
                    attrs_plus['id'] += '_{0}'.format(i)
                sub_ul_renderer = ChoiceFieldRenderer(name=self.name,
                                                      value=self.value,
                                                      attrs=attrs_plus,
                                                      choices=choice_label)
                sub_ul_renderer.choice_input_class = self.choice_input_class
                output.append(
                    format_html('<div class="checkbox">{0}{1}</div>',
                                choice_value, sub_ul_renderer.render()))
            else:
                # re-unpack the choice and handle HTML entities
                choice_value, choice_label = choice
                choice = (choice_value, mark_safe(
                    choice_label.encode('ascii', errors='xmlcharrefreplace')))
                w = self.choice_input_class(self.name, self.value,
                                            self.attrs.copy(), choice, i)
                output.append(format_html('<div class="checkbox">{0}</div>',
                                          smart_str(w)))
        output.append('</div>')
        return mark_safe('\n'.join(output))


class iCheckSelectMultiple(CheckboxSelectMultiple):
    renderer = iCheckFieldRenderer


class iCheckModelMultipleChoiceField(ModelMultipleChoiceField):
    widget = iCheckSelectMultiple


class PermissionForm(forms.ModelForm):
    """
    Generic form to be used to assign row-level permissions to an object.
    """
    widget = iCheckSelectMultiple
    staff_only = True
    max_checkboxes = 7  # Set based on number of rows visible in MVP layout

    def __init__(self, permission, *args, **kwargs):
        super(PermissionForm, self).__init__(*args, **kwargs)
        self.content_type = ContentType.objects.get_for_model(self.instance)
        self.permission = permission

        # determine if we should filter to only show staff
        user_queryset = get_user_model().objects.all()
        if self.staff_only:
            user_queryset = user_queryset.filter(is_staff=True)

        if user_queryset.count() <= self.max_checkboxes:
            user_widget = self.widget
        else:
            user_widget = SelectMultiple

        group_queryset = Group.objects.all()

        if group_queryset.count() <= self.max_checkboxes:
            group_widget = self.widget
        else:
            group_widget = SelectMultiple

        self.fields['users'] = ModelMultipleChoiceField(
            queryset=user_queryset,
            required=False,
            initial=self._users_with_perms,
            label_from_instance=lambda o: o.get_full_name(),
            widget=user_widget,
        )
        self.fields['users'].widget.attrs.setdefault('class', 'form-control')

        self.fields['groups'] = ModelMultipleChoiceField(
            queryset=group_queryset,
            required=False,
            initial=self._groups_with_perms,
            widget=group_widget,
        )
        self.fields['groups'].widget.attrs.setdefault('class', 'form-control')

    def __repr__(self):
        return "<%s: %s?%s>" % (
            self.__class__.__name__,
            self.permission.codename,
            urlencode({
                'staff_only': self.staff_only,
                'max_checkboxes': self.max_checkboxes,
            }))

    @cached_property
    def _users_with_perms(self):
        pks = UserObjectPermission.objects.filter(
            permission=self.permission,
            object_pk=self.instance.pk,
            content_type=self.content_type,
        ).values_list('user', flat=True)
        return get_user_model().objects.filter(pk__in=pks)

    @cached_property
    def _groups_with_perms(self):
        pks = GroupObjectPermission.objects.filter(
            permission=self.permission,
            object_pk=self.instance.pk,
            content_type=self.content_type,
        ).values_list('group', flat=True)
        return Group.objects.filter(pk__in=pks)

    def save(self, *args, **kwargs):
        if self.has_changed():
            # work out our before and after state
            users_before = self._users_with_perms
            users_after = self.cleaned_data['users']
            groups_before = self._groups_with_perms
            groups_after = self.cleaned_data['groups']

            # determine who has had their per object permissions rescinded
            rescind_users = set(users_before).difference(users_after)
            rescind_groups = set(groups_before).difference(groups_after)

            # determine who has had their per object permissions rescinded
            grant_users = set(users_after).difference(users_before)
            grant_groups = set(groups_after).difference(groups_before)

            # rescind and grant access
            for rescind in rescind_users.union(rescind_groups):
                remove_perm(self.permission.codename, rescind, self.instance)

            for grant in grant_users.union(grant_groups):
                assign_perm(self.permission.codename, grant, self.instance)

        return super(PermissionForm, self).save(*args, **kwargs)


class PermissionFormSetMixin(object):
    def __init__(self, instance, queryset, *args, **kwargs):
        super(PermissionFormSetMixin, self).__init__(*args, **kwargs)
        self.instance = instance
        self.queryset = queryset

    def total_form_count(self):
        return self.queryset.count()

    def _construct_form(self, i, **kwargs):
        return super(PermissionFormSetMixin, self)._construct_form(
            i, instance=self.instance, permission=self.queryset[i])

    def save(self, *args, **kwargs):
        res = []
        for form in self.forms:
            res.append(form.save())
        return res


def permissionformset_factory(model, staff_only=None, max_checkboxes=None):
    """
    For the specified Model, return a FormSet class which will allow you to
    control which specific users & groups should have the ability to change
    and delete a model instance.

    The resulting FormSet class will need to be instantiated with an instance
    of the appropriate type.

    The default behaviour is to grant permissions to staff users only.
    Specifying staff_only = False will result in a form that lists all users.

    When there is a large number of users a widget other than checkboxes might
    be preferred to allow for type-ahead searches. We use the Select2 jQuery
    plugin on to of a multiple select widget. Specify the integer number of
    users that will trigger the switch between checkboxes and Select2.

    :param model: Model
    :param staff_only: bool
    :param max_checkboxes: int
    :return: FormSet class
    """
    if staff_only is None:
        staff_only = PermissionForm.staff_only

    if max_checkboxes is None:
        max_checkboxes = PermissionForm.max_checkboxes

    meta = type('Meta', (), {'model': model, 'fields': ('id',)})
    form_class = type(
        '%sPermissionForm' % model.__name__,
        (PermissionForm,),
        {
            'Meta': meta,
            'staff_only': staff_only,
            'max_checkboxes': max_checkboxes,
        }
    )
    formset_base = formset_factory(form_class)
    formset_class = type('%sPermissionFormSet' % model.__name__,
                         (PermissionFormSetMixin, formset_base), {})
    return formset_class



from django.db.models.query import QuerySet
from django.utils import timezone


class DivisionQuerySet(QuerySet):
    """
    QuerySet to allow divisions to be easily filtered by their draft status.
    """
    def public(self):
        return self.filter(draft=False)


class StageQuerySet(QuerySet):

    def with_ladder(self):
        return self.filter(
            keep_ladder=True, ladder_summary__isnull=False).distinct()


class MatchQuerySet(QuerySet):

    def future(self, date=None):
        if date is None:
            date = timezone.now().date()
        return self.filter(date__gte=date)

    def playable(self):
        return self.exclude(is_bye=True)

    def videos(self):
        return self.filter(videos__isnull=False).order_by('datetime').distinct()


class StatisticQuerySet(QuerySet):

    def played(self):
        return self.exclude(played=0)



# -*- coding: utf-8 -*-

import logging
import re
from datetime import datetime
from decimal import Decimal

import numpy
from dateutil.relativedelta import relativedelta
from django.core.serializers import get_serializer
from django.db.models import Q
from django.shortcuts import get_object_or_404
from django.utils import timezone
from django.views.generic import dates

from tournamentcontrol.competition.models import LadderEntry, RankPoints, RankTeam, RankDivision

logger = logging.getLogger(__name__)


class NodeToContextMixin(object):

    def get_context_data(self, **kwargs):
        data = super(NodeToContextMixin, self).get_context_data(**kwargs)
        data['node'] = self.kwargs.get('node')
        return data


class IndexView(NodeToContextMixin, dates.ArchiveIndexView):
    model = RankPoints
    date_field = 'date'
    date_list_period = 'day'

    def get_date_list(self, *args, **kwargs):
        date_list = super(IndexView, self).get_date_list(*args, **kwargs)
        return date_list[:12]


class YearView(NodeToContextMixin, dates.YearArchiveView):
    model = RankPoints
    date_field = 'date'
    date_list_period = 'day'


class MonthView(NodeToContextMixin, dates.MonthArchiveView):
    model = RankPoints
    date_field = 'date'


class DayView(NodeToContextMixin, dates.DayArchiveView):
    model = RankPoints
    date_field = 'date'
    allow_future = True

    def get_context_data(self, **kwargs):
        data = super(DayView, self).get_context_data(**kwargs)
        division_list = (
            data['object_list']
            .values_list('team__division__title', 'team__division__slug')
            .distinct()
        )
        data['division_list'] = division_list
        return data


class DivisionView(NodeToContextMixin, dates.DayArchiveView):
    model = RankPoints
    date_field = 'date'
    template_name_suffix = '_detail'
    allow_future = True

    def get_queryset(self):
        queryset = super(DivisionView, self).get_queryset()
        slug = self.kwargs['slug']
        division_queryset = queryset.filter(team__division__slug=slug)
        return division_queryset.order_by('-points')

    def get_context_data(self, **kwargs):
        data = super(DivisionView, self).get_context_data(**kwargs)
        slug = self.kwargs['slug']
        data['object'] = get_object_or_404(RankDivision, slug=slug)
        return data


class Constants:
    ZERO = Decimal(0)
    ONE = Decimal(1)

    HALF = ONE / 2
    QUARTER = ONE / 4
    FIFTH = ONE / 5
    SIXTH = ONE / 6
    SEVENTH = ONE / 7
    EIGHTH = ONE / 8


def base(ladder_entry, win=15, draw=10, loss=5, forfeit_against=-20):
    return sum([
        ladder_entry.win * win,
        ladder_entry.draw * draw,
        ladder_entry.loss * loss,
        ladder_entry.forfeit_against * forfeit_against,
        win * 2 if ladder_entry.win and ladder_entry.margin > 15 else 0,
        win if ladder_entry.win and ladder_entry.margin > 10 else 0,
        win * Constants.HALF if ladder_entry.win and ladder_entry.margin > 5 else 0,
        loss * Constants.HALF if ladder_entry.loss and ladder_entry.margin < 2 else 0,
    ])


def no_decay(ladder_entry, at):
    return Constants.ONE


def rank(decay=no_decay, start=None, at=None, debug=None):
    """
    Return an `OrderedDict` of key values, where the key is a `RankDivision`
    and the value is a nested `OrderedDict` of weighted scores, keyed by a
    `RankTeam`.

    The optional `decay` callable is used to determine the amount of decay that
    is to be weighted on the contribution of an individual result. Default is
    to apply no decay.

    The optional `start` date specifies the earliest matches to be considered
    in establishing a ranking score. The default is to include all available
    matches.

    The optional `at` date specifies the point in time at which the rank is to
    be calculated. This will both allow for retrospectives (what was the rank
    on a certain date in the past) and what will the current rank decay to in
    the future (exclusive of any results that will occur in the interim). The
    default is "tomorrow" if not specified.

    Pass a regex club names to `debug` and any
    `competition.LadderEntry.team.club.title` that matches will emit some DEBUG
    level log messages about those entries.

    :param decay:
    :param start:
    :param at:
    :param debug:
    :return:
    """
    if start is None:
        start = LadderEntry.objects.earliest('match').match.date

    if at is None:
        at = timezone.now() + relativedelta(day=1)

    if isinstance(at, datetime):
        at = at.date()

    if debug is None:
        debug = r'^$'
    debug_re = re.compile(debug, re.I)

    ladder_entry_q = Q(match__date__gte=start, match__date__lt=at,
                       forfeit_for=0)
    ladder_entries = LadderEntry.objects.select_related(
        'match__stage__division__season__competition').order_by('match')

    table = {}
    for ladder_entry in ladder_entries.filter(ladder_entry_q):
        importance = ladder_entry.match.get_rank_importance()
        division = ladder_entry.team.get_rank_division()
        if importance is None or division is None:
            continue
        obj, __ = RankTeam.objects.get_or_create(club=ladder_entry.team.club, division=division)
        points = base(ladder_entry) * importance
        points_decay = points * decay(ladder_entry, at)
        if points_decay:
            team = table.setdefault(division, {}).setdefault(obj, {})
            team.setdefault('importance', []).append(importance)
            team.setdefault('points', []).append(points)
            team.setdefault('points_decay', []).append(points_decay)
            if debug_re.match(obj.club.title):
                logger.debug(
                    u'%r\t%-20s\t%-20s\twin=%s\tmargin=%s\t%s → %s\t%s',
                    ladder_entry.match,
                    ladder_entry.team,
                    ladder_entry.opponent,
                    ladder_entry.win,
                    ladder_entry.margin,
                    points,
                    points_decay,
                    importance)

    RankPoints.objects.filter(date__year=at.year, date__month=at.month).delete()

    for division in table:
        for team in table[division]:
            RankPoints.objects.create(
                team=team,
                points=numpy.mean(table[division][team]['points_decay']),
                date=at)

    return RankPoints.objects.filter(date=at)


def json_rank(stream=None, indent=4, *args, **kwargs):
    JSONSerializer = get_serializer('json')
    json_serializer = JSONSerializer()
    json_serializer.serialize(
        rank(*args, **kwargs),
        stream=stream,
        indent=indent,
        use_natural_foreign_keys=True,
        use_natural_primary_keys=True)
